<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines\Kernel;

use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\data_pipelines\Destination\ProcessingOperation;
use Drupal\data_pipelines\Destination\ProcessingOperationEnum;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Plugin\QueueWorker\DestinationWorker;

/**
 * Defines a class for testing queue indexing.
 *
 * @group data_pipelines
 *
 * @covers \Drupal\data_pipelines\EntityHandlers\DatasetStorage
 * @covers \Drupal\data_pipelines\Plugin\QueueWorker\DestinationWorker
 */
class DestinationQueueTest extends DatasetKernelTestBase {

  /**
   * Tests queue derivation.
   *
   * @covers \Drupal\data_pipelines\Plugin\Derivative\DatasetQueueWorkerDeriver
   */
  public function testQueueDerivation() {
    $queue = \Drupal::service('plugin.manager.queue_worker');
    assert($queue instanceof QueueWorkerManagerInterface);
    $this->assertCount(0, array_filter(array_keys($queue->getDefinitions()), function (string $id) {
      return strpos($id, 'data_pipelines_process') === 0;
    }));
    $dataset = $this->createTestDataset();
    $this->assertCount(1, array_filter(array_keys($queue->getDefinitions()), function (string $id) {
      return strpos($id, 'data_pipelines_process') === 0;
    }));
    $dataset->delete();
    $this->assertCount(0, array_filter(array_keys($queue->getDefinitions()), function (string $id) {
      return strpos($id, 'data_pipelines_process') === 0;
    }));
  }

  /**
   * Tests indexing.
   */
  public function testQueueWorker() {
    $destinationId = $this->randomMachineName();
    $dataset = $this->createTestDataset(['destinationId' => $destinationId]);
    $queue_id = $dataset->getProcessingQueueId();
    $queue_manager = \Drupal::service('plugin.manager.queue_worker');
    $process_queue = \Drupal::queue($queue_id);
    $dataset->queueProcessing($process_queue);
    // The state destination wants items one at a time.
    $this->assertEquals(4, $process_queue->numberOfItems());
    assert($queue_manager instanceof QueueWorkerManagerInterface);
    $this->assertTrue($queue_manager->hasDefinition($queue_id));
    $instance = $queue_manager->createInstance($queue_id);
    assert($instance instanceof DestinationWorker);
    $begin = $process_queue->claimItem()->data;
    assert($begin instanceof ProcessingOperation);
    $instance->processItem($begin);
    $this->assertEquals(ProcessingOperationEnum::BEGIN, $begin->getOperation());
    $this->assertEquals($destinationId, $begin->getDestinationId());
    $next = $process_queue->claimItem()->data;
    assert($next instanceof ProcessingOperation);
    $this->assertEquals($destinationId, $next->getDestinationId());
    $instance->processItem($next);
    $next = $process_queue->claimItem()->data;
    assert($next instanceof ProcessingOperation);
    $this->assertEquals($destinationId, $next->getDestinationId());
    $instance->processItem($next);
    $end = $process_queue->claimItem()->data;
    assert($end instanceof ProcessingOperation);
    $this->assertEquals($destinationId, $end->getDestinationId());
    $this->assertEquals(ProcessingOperationEnum::END, $end->getOperation());

    $data = \Drupal::state()->get('data_pipelines_test');
    $this->assertEquals([$dataset->id() => iterator_to_array($dataset->getDataIterator())], $data);
    $this->assertFalse($process_queue->claimItem());

    $expected = $data[$dataset->id()][0];

    $this->assertTrue($expected['should_we']);
    $this->assertEquals('bloggs, joe', $expected['full_name']);

    // Clean up.
    $dataset->delete();
    $data = \Drupal::state()->get('data_pipelines_test');
    $this->assertEmpty($data);
  }

  /**
   * Tests queue items are per destination.
   */
  public function testMultipleDestinations(): void {
    $id1 = $this->randomMachineName();
    $id2 = $this->randomMachineName();
    $destinations = [
      $this->createTestMemoryDestination(['id' => $id1]),
      $this->createTestMemoryDestination(['id' => $id2]),
    ];
    $dataset = $this->createTestDataset(['destinations' => $destinations]);
    $process_queue = \Drupal::queue($dataset->getProcessingQueueId());
    $dataset->queueProcessing($process_queue);

    // The state destination wants items one at a time and there is also an end
    // and begin task. So four items, and two destinations, so should be eight
    // in total.
    $this->assertEquals(8, $process_queue->numberOfItems());
  }

  /**
   * Tests cleanup queue creation.
   */
  public function testCleanup(): void {
    $destinationId = $this->randomMachineName();
    $file = $this->getTestFile(dirname(__DIR__, 1) . '/fixtures/test-pipeline-1-invalid.csv');
    $dataset = $this->createTestDataset([
      'source' => 'csv:file',
      'name' => $this->randomMachineName(),
      'machine_name' => mb_strtolower($this->randomMachineName()),
      'pipeline' => 'test_pipeline_1',
      'destinationId' => $destinationId,
      'csv_file' => $file,
      'invalid_values' => DatasetInterface::INVALID_REMOVE,
    ]);
    $queue_id = $dataset->getProcessingQueueId();
    $process_queue = \Drupal::queue($queue_id);
    $dataset->queueProcessing($process_queue);
    // The state destination wants items one at a time but two are invalid.
    // So we should have begin, process, end, cleanup totalling 4.
    $this->assertEquals(4, $process_queue->numberOfItems());
    $process_queue->claimItem();
    $process_queue->claimItem();
    $cleanup = $process_queue->claimItem()->data;
    assert($cleanup instanceof ProcessingOperation);
    $this->assertEquals(ProcessingOperationEnum::PROCESS_CLEANUP, $cleanup->getOperation());
    $this->assertEquals($destinationId, $cleanup->getDestinationId());
    $this->assertEquals([0, 1], $cleanup->getProcessingData());
  }

}
